/**********************************************************************
*
* ctxoxs.c
*
* Example Virtual Channel application
*
* Send and receive packets to/from the VDOXS virtual channel
*
* Copyright Citrix Systems, Inc.  All Rights Reserved.
*
* The code and comments contained in this file are provided "as is"
* without warranty of any kind, either expressed or implied,
* including, but not limited to, the implied warranties of
* merchantability and/or fitness for a particular purpose.
* You have the right to use, modify, reproduce, and distribute
* this file, and/or any modified version of this file, in any way
* that you find useful, provided that you agree that Citrix Systems
* has no warranty obligations or liability for this or any other
* sample file.
*
**********************************************************************/

/*
 *  Include files
 */
#include <windows.h> // Base headers
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tchar.h>


#include <wfapi.h>  // Citrix Server SDK

#include <vdoxs.h>  // Citrix shared header

#include "ctxoxs.h" // local header for CTXOXS


/*
 *  Global variables
 */
HANDLE      hVC           = NULL;
PCHAR       pPacket       = NULL;
static enum counter gridState[OXS_GRID_WIDTH][OXS_GRID_HEIGHT] = {
    {NONE, NONE, NONE}, {NONE, NONE, NONE}, {NONE, NONE, NONE}
};


/* Recursive check for a winning line */
static BOOLEAN recursiveCheckWin(enum counter c, int x, int y, int xinc, int yinc)
{
    if (x < 0 || y < 0 || x >= OXS_GRID_WIDTH || y >= OXS_GRID_HEIGHT)
        return TRUE;

    if (gridState[x][y] == NONE)
        return FALSE;

    if (gridState[x][y] != c)
        return FALSE;

    return recursiveCheckWin(c, x + xinc, y + yinc, xinc, yinc);
}


/* Check for any winning rows */
static BOOLEAN calculateWinner()
{
    int x, y, xinc, yinc;

    /* Horizontal wins */
    for (x = 0, y = 0, xinc = 1, yinc = 0; y < OXS_GRID_HEIGHT; y++)
    {
        if (recursiveCheckWin(gridState[x][y], x, y, xinc, yinc))
            goto winner;
    }

    /* Vertical wins */
    for (x = 0, y = 0, xinc = 0, yinc = 1; x < OXS_GRID_WIDTH; x++)
    {
        if (recursiveCheckWin(gridState[x][y], x, y, xinc, yinc))
            goto winner;
    }

    /* Top left to bottom right win */
    x = 0;
    y = 0;
    xinc = 1;
    yinc = 1;
    if (recursiveCheckWin(gridState[x][y], x, y, xinc, yinc)) 
        goto winner;

    /* Bottom left to top right win */
    x = 0;
    y = OXS_GRID_HEIGHT - 1;
    xinc = 1;
    yinc = -1;
    if (recursiveCheckWin(gridState[x][y], x, y, xinc, yinc))
        goto winner;
    
    return FALSE;

/* We have a winner. */
winner:
    {
        OXS_COORDINATE start;
        OXS_COORDINATE finish;

        start.x = x;
        start.y = y;
        finish.x = x + (xinc*(OXS_GRID_WIDTH - 1));
        finish.y = y + (yinc*(OXS_GRID_HEIGHT - 1));

        /* Tell the client to draw the winning line */
        SendWinningLinePacket(hVC, &start, &finish);
    }
    return TRUE;
}


/* Window message processing function */
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        case WM_MOVE:
            {
                /* Detect a window move and relay the new position */
                WINDOWINFO info;
                GetWindowInfo(hWnd, &info);

                SendMoveWindowPacket(hVC, info.rcClient.left, info.rcClient.top);
                return 0;
            }
        case WM_LBUTTONDOWN:
            {
                /* Detect a left mouse press on the game grid */
                static BOOLEAN endGame = FALSE;

                if (endGame == FALSE)
                {
                    OXS_COORDINATE c;

                    c.x = GET_X_LPARAM(lParam) / OXS_SQUARE_SIZE;
                    c.y = GET_Y_LPARAM(lParam) / OXS_SQUARE_SIZE;

                    if (gridState[c.x][c.y] == NONE)
                    {
                        static enum counter counter = CROSS;

                        /* Alternate between playing a nought and a cross */
                        if (counter == NOUGHT)
                            counter = CROSS;
                        else
                            counter = NOUGHT;

                        gridState[c.x][c.y] = counter;
                        SendPlayMovePacket(hVC, &c, counter);

                        /* Check if the game as been won */
                        endGame = calculateWinner();
                    }
                }
                return 0;
            }
        case WM_DESTROY:
            /* Detect destruction of server side window and notify client */
            SendCloseWindowPacket(hVC);
            PostQuitMessage(0);
            return 0;
        default:
            return DefWindowProc(hWnd, message, wParam, lParam);
    }
}


/****************************************************************************
 *
 *  WinMain
 *
 ***************************************************************************/

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nCmdShow)
{
    int         exitcode = SUCCESS;
    BOOLEAN     rc;
    LPTSTR      pSessionInfo  = NULL;
    PVDOXS_C2H  pVdQbuf       = NULL;
    PMODULE_C2H pmc2h         = NULL;
    PVD_C2H     pvdc2h        = NULL;
    DWORD       dwSize;
    DWORD       i;
    DWORD       ByteCount;
    MSG         message;
    HWND        hw;

    /* Determine if we are running in a session. */
    if ( WFQuerySessionInformation( WF_CURRENT_SERVER_HANDLE,
                            WF_CURRENT_SESSION,
                            WFSessionId,
                            &pSessionInfo, &ByteCount ) == FALSE ) {
        MessageBox(NULL,
                   _T("Noughts & Crosses example needs to run within a session."),
                   _T("No session"), MB_OK);
        exitcode = FAILURE;
        goto ExitMain;
    }
    else if ( (INT)(* pSessionInfo) == 0 ) {     // session 0 is console
        MessageBox(NULL,
                   _T("Noughts & Crosses example needs to run within a session."),
                   _T("No session"), MB_OK);
        exitcode = FAILURE;
        goto ExitMain;
    }

    /* Allocate data buffer the size of the largest packet to be sent */
    if ( (pPacket = (PCHAR) calloc(1, sizeof(OXS_HEADER) + sizeof(OXS_MOVE_WINDOW_PACKET))) == NULL ) {
        MessageBox(NULL,
           _T("Could not allocate memory for data buffer."),
           _T("Allocation Failure"), MB_OK);
        exitcode = FAILURE;
        goto ExitMain;
    }

    /* First, we want to open the channel and report the handle. */
    hVC = WFVirtualChannelOpen(WF_CURRENT_SERVER,
                               WF_CURRENT_SESSION,
                               CTXOXS_VIRTUAL_CHANNEL_NAME);

    if (!hVC) {
        MessageBox(NULL,
                   _T("Virtual Channel Open Failed."),
                   _T("Virtual Channel Open Failed"), MB_OK);
        exitcode = FAILURE;
        goto ExitMain;
    }

    /*  Now, query the channel to get data from the client. */
    rc = WFVirtualChannelQuery( hVC,
                                WFVirtualClientData,
                                &pVdQbuf,
                                &dwSize );

    if ( rc != TRUE ) {
        MessageBox(NULL,
                   _T("Make sure the client side $ICAROOT/config/module.ini ")
                   _T("is updated and the driver dll is in place."),
                   _T("Virtual Channel Query Failed"), MB_OK);
        exitcode = FAILURE;
        goto ExitMain;
    }

	/* Setup the Window */
	{
		WNDCLASSEX wc;
		TCHAR winClass[] = _T("oxs");
		TCHAR winTitle[] = _T("O and Xs");
        RECT clientArea = {0, 0, OXS_SQUARE_SIZE * OXS_GRID_WIDTH, OXS_SQUARE_SIZE * OXS_GRID_HEIGHT};
        DWORD winStyle = WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_VISIBLE;

		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = 0;
		wc.lpfnWndProc = WndProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInstance;
		wc.hIcon = NULL;
		wc.hCursor = LoadCursor(NULL, IDC_CROSS);
		wc.hbrBackground = (HBRUSH) (COLOR_BACKGROUND + 1);
		wc.lpszMenuName = NULL;
		wc.lpszClassName = winClass;
		wc.hIconSm = NULL;

		if (!RegisterClassEx(&wc))
			goto ExitMain;

        AdjustWindowRect(&clientArea, winStyle, FALSE);

        /* Create the server side game window */
		hw = CreateWindow(winClass,
                          winTitle,
						  winStyle,
                          CW_USEDEFAULT,
                          CW_USEDEFAULT,
                          clientArea.right - clientArea.left,
                          clientArea.bottom - clientArea.top,
                          NULL,
                          NULL,
                          hInstance,
                          NULL
                          );

        UpdateWindow(hw);
    }

   
    /* Send the window handle incase the application is seamless */
    SendParentWindowPacket(hVC, hw);

    {
        /* Send initial window information to position client window */
        WINDOWINFO info;     
        GetWindowInfo(hw, &info);

        SendMoveWindowPacket(hVC, info.rcClient.left, info.rcClient.top);
    }

    /* Win32 message loop */
    {
        BOOL rc;
        
        /* This will loop until the server side game window is closed */
        while ((rc = GetMessage(&message, hw, 0, 0)) != 0)
        {
            if (rc == -1)
            {
                break;
            }
            else
            {
                TranslateMessage(&message);
                DispatchMessage(&message);
            }
        }
    }

    /*  Finally, close the channel and free allocated memory. */
ExitMain:
    if ( pSessionInfo ) {
        WFFreeMemory( pSessionInfo );
    }
    if ( pVdQbuf ) {
        WFFreeMemory(pVdQbuf);
    }
    if ( hVC ) {
        WFVirtualChannelClose( hVC );
    }
    if ( pPacket ) {
        free( pPacket );
    }

    return exitcode;
}


/* Send parent window handle for use with seamless */
BOOLEAN SendParentWindowPacket(HANDLE hVC, HWND win)
{
    BOOLEAN     rc;
    ULONG ulen;
    POXS_HEADER h;
    POXS_PARENT_WINDOW_PACKET b;

    /* Add header */
    h = (POXS_HEADER) pPacket;
    h->type = OXS_PARENT_WINDOW;

    /* Add body */
    b = (POXS_PARENT_WINDOW_PACKET) (pPacket + sizeof(OXS_HEADER));
    b->parent = (UINT32) win;

    /*  Send the packet to the client. */
    rc = WFVirtualChannelWrite(hVC,
                               pPacket,
                               sizeof(OXS_HEADER) + sizeof(OXS_PARENT_WINDOW_PACKET),
                               &ulen);

	/* Assume perfect writes */
    if ( rc != TRUE ) {
        MessageBox(NULL,
           _T("Could not send Parent Window."),
           _T("Virtual Channel Write Failed"), MB_OK);
        return FALSE;
    }
	return TRUE;
}



/* Send packet to draw a counter */
BOOLEAN SendPlayMovePacket(HANDLE hVC, POXS_COORDINATE position, enum counter counter)
{
    BOOLEAN     rc;
    ULONG ulen;
    POXS_HEADER h;
    POXS_PLAY_MOVE_PACKET b;

    /* Add header */
    h = (POXS_HEADER) pPacket;
    h->type = OXS_PLAY_MOVE;

    /* Add body */
    b = (POXS_PLAY_MOVE_PACKET) (pPacket + sizeof(OXS_HEADER));
    b->position.x = position->x;
    b->position.y = position->y;
    b->counter = counter;

    /*  Send the packet to the client. */
    rc = WFVirtualChannelWrite( hVC,
                                pPacket,
                                sizeof(OXS_HEADER) + sizeof(OXS_PLAY_MOVE_PACKET),
                                &ulen );

	/* Assume perfect writes */
    if ( rc != TRUE ) {
        MessageBox(NULL,
           _T("Could not send Play Move."),
           _T("Virtual Channel Write Failed"), MB_OK);
        return FALSE;
    }
	return TRUE;
}


BOOLEAN SendWinningLinePacket(HANDLE hVC, POXS_COORDINATE start, POXS_COORDINATE finish)
{
    BOOLEAN     rc;
    ULONG ulen;
    POXS_HEADER h;
    POXS_WINNING_LINE_PACKET b;

    /* Add header */
    h = (POXS_HEADER) pPacket;
    h->type = OXS_WINNING_LINE;

    /* Add body */
    b = (POXS_WINNING_LINE_PACKET) (pPacket + sizeof(OXS_HEADER));
    b->startpos.x = start->x;
    b->startpos.y = start->y;
    b->finishpos.x = finish->x;
    b->finishpos.y = finish->y;

    /*  Send the packet to the client. */
    rc = WFVirtualChannelWrite(hVC,
                               pPacket,
                               sizeof(OXS_HEADER) + sizeof(OXS_WINNING_LINE_PACKET),
                               &ulen);

	/* Assume perfect writes */
    if ( rc != TRUE ) {
        MessageBox(NULL,
           _T("Could not send Winning Line."),
           _T("Virtual Channel Write Failed"), MB_OK);
        return FALSE;
    }
	return TRUE;
}


BOOLEAN SendMoveWindowPacket(HANDLE hVC, UINT32 x, UINT32 y)
{
    BOOLEAN     rc;
    ULONG ulen;
    POXS_HEADER h;
    POXS_MOVE_WINDOW_PACKET b;

    /* Add header */
    h = (POXS_HEADER) pPacket;
    h->type = OXS_MOVE_WINDOW;

    /* Add body */
    b = (POXS_MOVE_WINDOW_PACKET) (pPacket + sizeof(OXS_HEADER));
    b->x = x;
    b->y = y;

    /*  Send the packet to the client. */
    rc = WFVirtualChannelWrite(hVC,
                               pPacket,
                               sizeof(OXS_HEADER) + sizeof(OXS_MOVE_WINDOW_PACKET),
                               &ulen);

	/* Assume perfect writes */
    if ( rc != TRUE ) {
        MessageBox(NULL,
           _T("Could not send Move Window."),
           _T("Virtual Channel Write Failed"), MB_OK);
        return FALSE;
    }
	return TRUE;
}


BOOLEAN SendCloseWindowPacket(HANDLE hVC)
{
    BOOLEAN     rc;
    ULONG ulen;
    POXS_HEADER h;

    /* Add header */
    h = (POXS_HEADER) pPacket;
    h->type = OXS_CLOSE_WINDOW;

    /*  Send the packet to the client. */
    rc = WFVirtualChannelWrite(hVC,
                               pPacket,
                               sizeof(OXS_HEADER),
                               &ulen);

    	/* Assume perfect writes */
    if ( rc != TRUE ) {
        MessageBox(NULL,
           _T("Could not send Close Window."),
           _T("Virtual Channel Write Failed"), MB_OK);
        return FALSE;
    }
	return TRUE;
}
